@@ -1,4 +1,6 @@
 """streaming_pileupy"""
+import itertools
+
 __author__ = """Warren W. Kretzschmar"""
 __email__ = "winni@warrenwk.com"
 __version__ = "0.5.4"
@@ -48,7 +50,7 @@ def get_rg_lookup_table(header):
 class MpileupWriter:
     out_fh: IO[str]
     samples: List[str]
-    buffer: DefaultDict[int, DefaultDict[str, List[Tuple[str, int]]]]
+    buffer: DefaultDict[int, List[Tuple[str, str, int]]]
     ref_bases: Dict[int, str]
     _chrom: str
     min_bq: int
@@ -58,17 +60,18 @@ class MpileupWriter:
         return cls(
             out_fh=fh,
             samples=samples,
-            buffer=defaultdict(lambda: defaultdict(list)),
+            buffer=defaultdict(list),
             ref_bases={},
             _chrom="",
             min_bq=0,
         )
 
     def add_base(self, sample: str, pos: int, base: str, qual: int, ref: str) -> None:
-        self.buffer[pos][sample].append((base, qual))
         self.ref_bases[pos] = ref
+        buffer_pos_sample = self.buffer[pos]
         if qual < self.min_bq:
-            self.buffer[pos][sample].pop()
+            return
+        buffer_pos_sample.append((sample, base, qual))
 
     @property
     def chrom(self):
@@ -93,27 +96,31 @@ class MpileupWriter:
     def _flush_position(self, pos):
         ref_base = self.ref_bases.pop(pos)
         pos_bases = self.buffer.pop(pos)
-        outline = [f"{self.chrom}\t{pos + 1}\t{ref_base.upper()}"]
-        outline += list(self._get_sample_strings(pos_bases))
-        outline.append("\n")
-        self.out_fh.write("".join(outline))
+        out_line = itertools.chain(
+            [f"{self.chrom}\t{pos + 1}\t{ref_base.upper()}"],
+            self._get_sample_strings(pos_bases),
+            ["\n"],
+        )
+        self.out_fh.write("".join(out_line))
 
     def _get_sample_strings(self, pos_bases):
+        samp_bases = defaultdict(list)
+        samp_quals = defaultdict(list)
+        for sample, base, qual in pos_bases:
+            samp_bases[sample].append(base)
+            samp_quals[sample].append(qual)
         for sample in self.samples:
-            if sample in pos_bases:
-                sample_data = pos_bases[sample]
-            else:
-                sample_data = None
-            if not sample_data:
-                yield "\t0\t*\t*"
-            else:
-                bases, quals = list(zip(*sample_data))
+            if sample in samp_bases:
+                bases = samp_bases[sample]
+                quals = samp_quals[sample]
                 quals = [chr(q + 33) for q in quals]
                 assert len(bases) == len(quals)
                 n_bases = len(bases)
                 bases = "".join(bases)
                 quals = "".join(quals)
                 yield f"\t{n_bases}\t{bases}\t{quals}"
+            else:
+                yield "\t0\t*\t*"
 
 
 @dataclass
@@ -144,10 +151,11 @@ class RecordReader:
         if logger.isEnabledFor(logging.DEBUG):
             log_record_info(match_base, rec, sequence)
 
+        read_group = self.rg_table[rg_id]
         for read_pos, ref_pos, ref_base in rec.get_aligned_pairs(
             matches_only=True, with_seq=True
         ):
-            logger.debug("read: %s, ref: %s, ref_base: %s", read_pos, ref_pos, ref_base)
+            # logger.debug("read: %s, ref: %s, ref_base: %s", read_pos, ref_pos, ref_base)
 
             base = sequence[read_pos].upper()
             if base == ref_base.upper():
@@ -156,7 +164,7 @@ class RecordReader:
                 base = base.lower()
 
             self.writer.add_base(
-                sample=self.rg_table[rg_id],
+                sample=read_group,
                 pos=ref_pos,
                 base=base,
                 qual=qualities[read_pos],
